import math
from typing import Union

from . import *


class Usage(CustomIdentified):

    def __init__(self, entity: str, *, identity: str = None,
                 type_uri: str = PROV_USAGE) -> None:
        super().__init__(identity=identity, type_uri=type_uri)
        self.entity = URIProperty(self, PROV_ENTITY, 1, 1,
                                  initial_value=entity)
        self.roles = URIProperty(self, PROV_ROLES, 0, math.inf)
        self.validate()


def build_usage(identity: str, *, type_uri: str = PROV_USAGE) -> SBOLObject:
    missing = PYSBOL3_MISSING
    obj = Usage(missing, identity=identity, type_uri=type_uri)
    # Remove the dummy values
    obj._properties[PROV_ENTITY] = []
    return obj


Document.register_builder(PROV_USAGE, build_usage)


class Agent(CustomTopLevel):

    def __init__(self, identity: str, *, type_uri: str = PROV_AGENT) -> None:
        super().__init__(identity, type_uri)
        self.validate()


Document.register_builder(PROV_AGENT, Agent)


class Plan(CustomTopLevel):

    def __init__(self, identity: str, *, type_uri: str = PROV_PLAN) -> None:
        super().__init__(identity, type_uri)
        self.validate()


Document.register_builder(PROV_PLAN, Plan)


class Association(CustomIdentified):

    def __init__(self, agent: Union[str, Identified],
                 *, identity: str = None,
                 type_uri: str = PROV_ASSOCIATION) -> None:
        super().__init__(identity=identity, type_uri=type_uri)
        self.roles = URIProperty(self, PROV_ROLES, 0, math.inf)
        self.plan = ReferencedObject(self, PROV_PLANS, 0, 1)
        self.agent = ReferencedObject(self, PROV_AGENTS, 1, 1,
                                      initial_value=agent)
        self.validate()


def build_association(identity: str, *, type_uri: str = PROV_USAGE) -> SBOLObject:
    missing = PYSBOL3_MISSING
    obj = Association(missing, identity=identity, type_uri=type_uri)
    # Remove the dummy values
    obj._properties[PROV_AGENTS] = []
    return obj


Document.register_builder(PROV_ASSOCIATION, build_association)


class Activity(CustomTopLevel):

    def __init__(self, identity: str, *, type_uri: str = PROV_ACTIVITY) -> None:
        super().__init__(identity, type_uri)
        self.types = URIProperty(self, SBOL_TYPE, 0, math.inf)
        self.start_time = DateTimeProperty(self, PROV_STARTED_AT_TIME, 0, 1)
        self.end_time = DateTimeProperty(self, PROV_ENDED_AT_TIME, 0, 1)
        self.usage = OwnedObject(self, PROV_QUALIFIED_USAGE, 0, math.inf,
                                 type_constraint=Usage)
        self.association = OwnedObject(self, PROV_QUALIFIED_ASSOCIATION, 0, math.inf,
                                       type_constraint=Association)
        self.validate()


Document.register_builder(PROV_ACTIVITY, Activity)



import abc
from typing import Union

from . import *

from .om_unit import Unit


class CompoundUnit(Unit, abc.ABC):

    def __init__(self, identity: str, symbol: str, label: str,
                 type_uri: str) -> None:
        super().__init__(identity, symbol, label, type_uri)


class UnitMultiplication(CompoundUnit):

    def __init__(self, identity: str, symbol: str, label: str,
                 term1: Union[Unit, str], term2: Union[Unit, str],
                 *, type_uri: str = OM_UNIT_MULTIPLICATION) -> None:
        super().__init__(identity, symbol, label, type_uri)
        self.term1 = ReferencedObject(self, OM_HAS_TERM1, 1, 1,
                                      initial_value=term1)
        self.term2 = ReferencedObject(self, OM_HAS_TERM2, 1, 1,
                                      initial_value=term2)
        self.validate()


def build_unit_multiplication(identity: str,
                              *, type_uri: str = OM_UNIT_MULTIPLICATION) -> SBOLObject:
    missing = PYSBOL3_MISSING
    obj = UnitMultiplication(identity, missing, missing, missing, missing,
                             type_uri=type_uri)
    # Remove the dummy values
    obj._properties[OM_SYMBOL] = []
    obj._properties[OM_LABEL] = []
    obj._properties[OM_HAS_TERM1] = []
    obj._properties[OM_HAS_TERM2] = []
    return obj


Document.register_builder(OM_UNIT_MULTIPLICATION, build_unit_multiplication)


class UnitDivision(CompoundUnit):

    def __init__(self, identity: str, symbol: str, label: str,
                 numerator: Union[Unit, str], denominator: Union[Unit, str],
                 *, type_uri: str = OM_UNIT_DIVISION) -> None:
        super().__init__(identity, symbol, label, type_uri)
        self.numerator = ReferencedObject(self, OM_HAS_NUMERATOR, 1, 1,
                                          initial_value=numerator)
        self.denominator = ReferencedObject(self, OM_HAS_DENOMINATOR, 1, 1,
                                            initial_value=denominator)
        self.validate()


def build_unit_division(identity: str,
                        *, type_uri: str = OM_UNIT_DIVISION) -> SBOLObject:
    missing = PYSBOL3_MISSING
    obj = UnitDivision(identity, missing, missing, missing, missing, type_uri=type_uri)
    # Remove the dummy values
    obj._properties[OM_SYMBOL] = []
    obj._properties[OM_LABEL] = []
    obj._properties[OM_HAS_NUMERATOR] = []
    obj._properties[OM_HAS_DENOMINATOR] = []
    return obj


Document.register_builder(OM_UNIT_DIVISION, build_unit_division)


class UnitExponentiation(CompoundUnit):

    def __init__(self, identity: str, symbol: str, label: str,
                 base: Union[Unit, str], exponent: int,
                 *, type_uri: str = OM_UNIT_EXPONENTIATION) -> None:
        super().__init__(identity, symbol, label, type_uri)
        self.base = ReferencedObject(self, OM_HAS_BASE, 1, 1,
                                     initial_value=base)
        self.exponent = IntProperty(self, OM_HAS_EXPONENT, 1, 1,
                                    initial_value=exponent)
        self.validate()


def build_unit_exponentiation(identity: str,
                              *, type_uri: str = OM_UNIT_EXPONENTIATION) -> SBOLObject:
    missing = PYSBOL3_MISSING
    obj = UnitExponentiation(identity, missing, missing, missing, 1, type_uri=type_uri)
    # Remove the dummy values
    obj._properties[OM_SYMBOL] = []
    obj._properties[OM_LABEL] = []
    obj._properties[OM_HAS_BASE] = []
    obj._properties[OM_HAS_EXPONENT] = []
    return obj


Document.register_builder(OM_UNIT_EXPONENTIATION, build_unit_exponentiation)



from typing import Union, Any, List, Optional

import rdflib

from . import *


class URIPropertyMixin:

    @staticmethod
    def to_user(value: Any) -> str:
        return str(value)

    @staticmethod
    def from_user(value: Any) -> rdflib.URIRef:
        if not isinstance(value, str):
            raise TypeError(f'Expecting string, got {type(value)}')
        return rdflib.URIRef(value)


class URISingletonProperty(URIPropertyMixin, SingletonProperty):

    def __init__(self, property_owner: Any, property_uri: str,
                 lower_bound: int, upper_bound: int,
                 validation_rules: Optional[List] = None,
                 initial_value: Optional[str] = None):
        super().__init__(property_owner, property_uri,
                         lower_bound, upper_bound, validation_rules)
        if initial_value:
            self.set(initial_value)


class URIListProperty(URIPropertyMixin, ListProperty):

    def __init__(self, property_owner: Any, property_uri: str,
                 lower_bound: int, upper_bound: int,
                 validation_rules: Optional[List] = None,
                 initial_value: Optional[str] = None):
        super().__init__(property_owner, property_uri,
                         lower_bound, upper_bound, validation_rules)
        if initial_value:
            self.set(initial_value)


def URIProperty(property_owner: Any, property_uri: str,
                lower_bound: int, upper_bound: Union[int, float],
                *,  # require keywords from here
                validation_rules: Optional[List] = None,
                initial_value: Optional[Union[str, List[str]]] = None) -> Property:
    if upper_bound == 1:
        return URISingletonProperty(property_owner, property_uri,
                                    lower_bound, upper_bound,
                                    validation_rules, initial_value)
    return URIListProperty(property_owner, property_uri,
                           lower_bound, upper_bound,
                           validation_rules, initial_value)



from typing import Union, Any, List, Optional

import rdflib

from . import *


class ReferencedURI(str):

    def __eq__(self, other):
        return super().__eq__(other)

    def lookup(self):
        if hasattr(self, 'parent'):
            return self.parent.document.find(str(self))
        else:
            # TODO: Should the lack of a parent raise an error?
            return None


class ReferencedObjectMixin:

    def to_user(self, value: Any) -> str:
        result = ReferencedURI(str(value))
        if hasattr(self, 'property_owner'):
            parent = self.property_owner
            result.parent = parent
        return result

    @staticmethod
    def from_user(value: Any) -> rdflib.URIRef:
        if isinstance(value, SBOLObject):
            value = value.identity
        if not isinstance(value, str):
            raise TypeError(f'Expecting string, got {type(value)}')
        return rdflib.URIRef(value)

    def maybe_add_to_document(self, value: Any) -> None:
        # if not isinstance(value, TopLevel):
        #     return
        if hasattr(self, 'property_owner'):
            if self.property_owner and self.property_owner.document:
                try:
                    self.property_owner.document.add(value)
                except TypeError:
                    # Not an appropriate type to be added, ignore
                    pass


class ReferencedObjectSingleton(ReferencedObjectMixin, SingletonProperty):

    def __init__(self, property_owner: Any, property_uri: str,
                 lower_bound: int, upper_bound: int,
                 validation_rules: Optional[List] = None,
                 initial_value: Optional[str] = None):
        super().__init__(property_owner, property_uri,
                         lower_bound, upper_bound, validation_rules)
        if initial_value:
            self.set(initial_value)

    def set(self, value: Any) -> None:
        super().set(value)
        self.maybe_add_to_document(value)


class ReferencedObjectList(ReferencedObjectMixin, ListProperty):

    def __init__(self, property_owner: Any, property_uri: str,
                 lower_bound: int, upper_bound: int,
                 validation_rules: Optional[List] = None,
                 initial_value: Optional[str] = None):
        super().__init__(property_owner, property_uri,
                         lower_bound, upper_bound, validation_rules)
        if initial_value:
            self.set(initial_value)

    def item_added(self, item: Any) -> None:
        self.maybe_add_to_document(item)


def ReferencedObject(property_owner: Any, property_uri: str,
                     lower_bound: int, upper_bound: Union[int, float],
                     validation_rules: Optional[List] = None,
                     initial_value: Optional[Union[str, List[str]]] = None) -> Property:
    if upper_bound == 1:
        return ReferencedObjectSingleton(property_owner, property_uri,
                                         lower_bound, upper_bound,
                                         validation_rules, initial_value)
    return ReferencedObjectList(property_owner, property_uri,
                                lower_bound, upper_bound,
                                validation_rules, initial_value)



import abc
import math
from typing import Union

from . import *
from .om_prefix import Prefix


class Unit(CustomTopLevel, abc.ABC):
    """om:Unit is an abstract base class.

    See Appendix A Section A.2 of the SBOL 3.0 specificiation.
    """

    def __init__(self, identity: str, symbol: str, label: str,
                 type_uri: str) -> None:
        super().__init__(identity, type_uri)
        self.symbol = TextProperty(self, OM_SYMBOL, 1, 1,
                                   initial_value=symbol)
        self.alternative_symbols = TextProperty(self, OM_ALTERNATIVE_SYMBOL,
                                                0, math.inf)
        self.label = TextProperty(self, OM_LABEL, 1, 1,
                                  initial_value=label)
        self.alternative_labels = TextProperty(self, OM_ALTERNATIVE_LABEL,
                                               0, math.inf)
        self.comment = TextProperty(self, OM_COMMENT, 0, 1)
        self.long_comment = TextProperty(self, OM_LONG_COMMENT, 0, 1)


class Measure(CustomIdentified):

    def __init__(self, value: float, unit: str,
                 *, identity: str = None,
                 type_uri: str = OM_MEASURE) -> None:
        super().__init__(identity=identity, type_uri=type_uri)
        self.value = FloatProperty(self, OM_HAS_NUMERICAL_VALUE, 1, 1,
                                   initial_value=value)
        self.types = URIProperty(self, SBOL_TYPE, 0, math.inf)
        self.unit = URIProperty(self, OM_HAS_UNIT, 1, 1,
                                initial_value=unit)
        self.validate()


def build_measure(identity: str, *, type_uri: str = OM_MEASURE) -> SBOLObject:
    missing = PYSBOL3_MISSING
    obj = Measure(1.0, missing, identity=identity, type_uri=type_uri)
    # Remove the dummy values
    obj._properties[OM_HAS_NUMERICAL_VALUE] = []
    obj._properties[OM_HAS_UNIT] = []
    return obj


Document.register_builder(OM_MEASURE, build_measure)


class SingularUnit(Unit):

    def __init__(self, identity: str, symbol: str, label: str,
                 *, type_uri: str = OM_SINGULAR_UNIT) -> None:
        super().__init__(identity, symbol, label, type_uri)
        self.unit = ReferencedObject(self, OM_HAS_UNIT, 0, 1)
        self.factor = FloatProperty(self, OM_HAS_FACTOR, 0, 1)
        self.validate()


def build_singular_unit(identity: str,
                        *, type_uri: str = OM_SINGULAR_UNIT) -> SBOLObject:
    missing = PYSBOL3_MISSING
    obj = SingularUnit(identity, missing, missing, type_uri=type_uri)
    # Remove the dummy values
    obj._properties[OM_SYMBOL] = []
    obj._properties[OM_LABEL] = []
    return obj


Document.register_builder(OM_SINGULAR_UNIT, build_singular_unit)


class PrefixedUnit(Unit):

    def __init__(self, identity: str, symbol: str, label: str,
                 unit: Union[Unit, str],
                 prefix: Union[Prefix, str],
                 *, type_uri: str = OM_PREFIXED_UNIT) -> None:
        super().__init__(identity, symbol, label, type_uri)
        self.unit = ReferencedObject(self, OM_HAS_UNIT, 1, 1,
                                     initial_value=unit)
        self.prefix = ReferencedObject(self, OM_HAS_PREFIX, 1, 1,
                                       initial_value=prefix)
        self.validate()


def build_prefixed_unit(identity: str,
                        *, type_uri: str = OM_PREFIXED_UNIT) -> SBOLObject:
    missing = PYSBOL3_MISSING
    obj = PrefixedUnit(identity, missing, missing, missing, missing, type_uri=type_uri)
    # Remove the dummy values
    obj._properties[OM_SYMBOL] = []
    obj._properties[OM_LABEL] = []
    obj._properties[OM_HAS_UNIT] = []
    obj._properties[OM_HAS_PREFIX] = []
    return obj


Document.register_builder(OM_PREFIXED_UNIT, build_prefixed_unit)



import unittest

import sbol3


class TestSequenceFeature(unittest.TestCase):

    def test_create(self):
        es1 = sbol3.EntireSequence(sbol3.PYSBOL3_MISSING)
        locations = [es1]
        sf = sbol3.SequenceFeature(locations)
        self.assertEqual(locations, sf.locations)

    def test_validation(self):
        locations = []
        with self.assertRaises(sbol3.ValidationError):
            sbol3.SequenceFeature(locations)


if __name__ == '__main__':
    unittest.main()



from . import *


class Attachment(TopLevel):

    def __init__(self, identity: str, source: str,
                 *, type_uri: str = SBOL_ATTACHMENT):
        super().__init__(identity, type_uri)
        self.source = URIProperty(self, SBOL_SOURCE, 1, 1,
                                  initial_value=source)
        self.format = URIProperty(self, SBOL_FORMAT, 0, 1)
        self.size = IntProperty(self, SBOL_SIZE, 0, 1)
        self.hash = TextProperty(self, SBOL_HASH, 0, 1)
        self.hash_algorithm = TextProperty(self, SBOL_HASH_ALGORITHM, 0, 1)
        self.validate()

    def validate(self) -> None:
        super().validate()
        # An Attachment must have 1 source
        if self.source is None:
            message = f'Attachment {self.identity} must have a source'
            raise ValidationError(message)


def build_attachment(identity: str, *, type_uri: str = SBOL_COMPONENT) -> SBOLObject:
    missing = PYSBOL3_MISSING
    obj = Attachment(identity, missing, type_uri=type_uri)
    # Remove the dummy values
    obj._properties[SBOL_SOURCE] = []
    return obj


Document.register_builder(SBOL_ATTACHMENT, build_attachment)



from setuptools import setup

setup(name='sbol3',
      version='1.0a5',
      description='Python implementation of SBOL 3 standard',
      python_requires='>=3.6',
      url='https://github.com/SynBioDex/pySBOL3',
      author='Bryan Bartley',
      author_email='editors@sbolstandard.org',
      license='Apache-2',
      # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
            # How mature is this project? Common values are
            #   3 - Alpha
            #   4 - Beta
            #   5 - Production/Stable
            'Development Status :: 3 - Alpha',

            # Indicate who your project is intended for
            'Intended Audience :: Developers',

            # Pick your license as you wish (should match "license" above)
            'License :: OSI Approved :: Apache Software License',

            # Specify the Python versions you support here. In particular, ensure
            # that you indicate whether you support Python 2, Python 3 or both.
            'Programming Language :: Python :: 3'
      ],
      # What does your project relate to?
      keywords='synthetic biology',
      packages=['sbol3'],
      install_requires=[
            'rdflib>=5.0',
            'rdflib-jsonld',
            'python-dateutil'
      ],
      test_suite='test',
      tests_require=[
            'pycodestyle>=2.6.0'
      ])



import abc
import math

from . import *


class Feature(Identified, abc.ABC):
    """Feature is an abstract base class."""

    def __init__(self, identity: str, type_uri: str) -> None:
        super().__init__(identity, type_uri)
        self.roles = URIProperty(self, SBOL_ROLE, 0, math.inf)
        self.orientation = URIProperty(self, SBOL_ORIENTATION, 0, 1)

    def validate(self) -> None:
        super().validate()
        # If there is an orientation, it must be in the valid set
        if self.orientation is not None:
            valid_orientations = [SBOL_INLINE, SBOL_REVERSE_COMPLEMENT]
            if self.orientation not in valid_orientations:
                message = f'{self.orientation} is not a valid orientation'
                raise ValidationError(message)



from . import *


class ExperimentalData(TopLevel):

    def __init__(self, identity: str, *, type_uri: str = SBOL_EXPERIMENTAL_DATA):
        super().__init__(identity, type_uri)


Document.register_builder(SBOL_EXPERIMENTAL_DATA, ExperimentalData)



import math
from typing import Union

from . import *
# Feature is not exported
from .feature import Feature


class SubComponent(Feature):

    def __init__(self, instance_of: Union[Identified, str],
                 *, identity: str = None, type_uri: str = SBOL_SUBCOMPONENT) -> None:
        super().__init__(identity, type_uri)
        self.role_integration = URIProperty(self, SBOL_ROLE, 0, 1)
        self.instance_of = ReferencedObject(self, SBOL_INSTANCE_OF, 1, 1,
                                            initial_value=instance_of)
        self.source_locations = OwnedObject(self, SBOL_SOURCE_LOCATION, 0, math.inf,
                                            type_constraint=Location)
        self.locations = OwnedObject(self, SBOL_LOCATION, 0, math.inf,
                                     type_constraint=Location)
        self.validate()

    def validate(self) -> None:
        super().validate()
        # If there is an orientation, it must be in the valid set
        if not self.instance_of:
            raise ValidationError('SubComponent must have an instance_of')


def build_subcomponent(identity: str, type_uri: str = SBOL_SUBCOMPONENT) -> Identified:
    """Used by Document to construct a SubComponent when reading an SBOL file.
    """
    missing = PYSBOL3_MISSING
    obj = SubComponent(missing, identity=identity, type_uri=type_uri)
    obj._properties[SBOL_INSTANCE_OF] = []
    return obj


Document.register_builder(SBOL_SUBCOMPONENT, build_subcomponent)



from typing import Union, List

from .property_base import Property

# URIProperty typing
uri_list = Union[List[str], Property]
uri_singleton = Union[str, Property]
# Owned object property
oo_list = Union[List['Identified'], Property]



import os
import unittest

import sbol3

MODULE_LOCATION = os.path.dirname(os.path.abspath(__file__))
SBOL3_LOCATION = os.path.join(MODULE_LOCATION, 'SBOLTestSuite', 'SBOL3')


class TestSIPrefix(unittest.TestCase):

    def test_create(self):
        display_id = 'si_prefix'
        symbol = 'kilo'
        label = 'Kilo'
        factor = 1000
        prefix = sbol3.SIPrefix(display_id, symbol, label, factor)
        self.assertIsNotNone(prefix)
        self.assertEqual(symbol, prefix.symbol)
        self.assertEqual(label, prefix.label)
        self.assertAlmostEqual(float(factor), prefix.factor, 10)
        self.assertIsInstance(prefix.factor, float)

    def test_read_from_file(self):
        test_file = os.path.join(SBOL3_LOCATION, 'measurement_entity',
                                 'measurement', 'measurement.nt')
        doc = sbol3.Document()
        doc.read(test_file, sbol3.NTRIPLES)
        uri = 'https://sbolstandard.org/examples/milli'
        si_prefix = doc.find(uri)
        self.assertIsNotNone(si_prefix)
        self.assertIsInstance(si_prefix, sbol3.SIPrefix)
        self.assertEqual('m', si_prefix.symbol)
        self.assertEqual('milli', si_prefix.label)
        self.assertCountEqual(['m1', 'm2'], si_prefix.alternative_symbols)
        self.assertCountEqual(['milli1', 'milli2'], si_prefix.alternative_labels)
        self.assertAlmostEqual(0.001, si_prefix.factor, 3)
        self.assertIsNotNone(si_prefix.long_comment)
        self.assertIsNotNone(si_prefix.description)
        self.assertIsNotNone(si_prefix.comment)


class TestBinaryPrefix(unittest.TestCase):

    def test_create(self):
        display_id = 'binary_prefix'
        symbol = 'kilo'
        label = 'Kilo'
        factor = 1000
        prefix = sbol3.BinaryPrefix(display_id, symbol, label, factor)
        self.assertIsNotNone(prefix)
        self.assertEqual(symbol, prefix.symbol)
        self.assertEqual(label, prefix.label)
        self.assertAlmostEqual(float(factor), prefix.factor, 10)
        self.assertIsInstance(prefix.factor, float)

    def test_read_from_file(self):
        # BinaryPrefix does not appear in any test files
        pass


if __name__ == '__main__':
    unittest.main()



from typing import Union, Any, List, Optional

import rdflib

from . import *


class FloatPropertyMixin:

    def from_user(self, value: Union[float, int, None]) -> Union[None, rdflib.Literal]:
        if value is None:
            return None
        if isinstance(value, int):
            value = float(value)
        if not isinstance(value, float):
            raise TypeError(f'Expecting float, got {type(value)}')
        return rdflib.Literal(value)

    def to_user(self, value: Any) -> float:
        return float(value)


class FloatSingletonProperty(FloatPropertyMixin, SingletonProperty):

    def __init__(self, property_owner: Any, property_uri: str,
                 lower_bound: int, upper_bound: int,
                 validation_rules: Optional[List] = None,
                 initial_value: Union[float, int, None] = None):
        super().__init__(property_owner, property_uri,
                         lower_bound, upper_bound, validation_rules)
        if initial_value is not None:
            self.set(initial_value)


def FloatProperty(property_owner: Any, property_uri: str,
                  lower_bound: int, upper_bound: Union[int, float],
                  validation_rules: Optional[List] = None,
                  initial_value: Optional[Union[float, List[float]]] = None) -> Property:
    if upper_bound == 1:
        return FloatSingletonProperty(property_owner, property_uri,
                                      lower_bound, upper_bound,
                                      validation_rules, initial_value)
    raise ValueError('Upper bound > 1 not handled yet')



import posixpath
import uuid
from collections import defaultdict
from urllib.parse import urlparse
from typing import Dict, Callable, Optional

from . import *


class SBOLObject:

    def __init__(self, name: str, type_uri: str) -> None:
        self._properties = defaultdict(list)
        self._owned_objects = defaultdict(list)
        # Does this need to be a property? It does not get serialized to the RDF file.
        # Could it be an attribute that gets composed on the fly? Keep it simple for
        # now, and change to a property in the future if needed.
        self._identity = SBOLObject._make_identity(name)
        self.document = None
        # This is the URI for this object's type, used for serialization
        self.type_uri = type_uri

    def __setattr__(self, name, value):
        try:
            self.__dict__[name].set(value)
        except AttributeError:
            # Attribute set does not exist
            object.__setattr__(self, name, value)
        except KeyError:
            # property name does not exist
            object.__setattr__(self, name, value)

    def __getattribute__(self, name):
        # Call the default method
        result = object.__getattribute__(self, name)
        if hasattr(result, '_sbol_singleton'):
            result = result.get()
        return result

    def _validate_identity(self) -> None:
        # TODO: identity must be a URI
        # TODO: can identity be None?
        pass

    @staticmethod
    def _is_url(name: str) -> bool:
        parsed = urlparse(name)
        return bool(parsed.scheme and parsed.netloc and parsed.path)

    @staticmethod
    def _make_identity(name: str) -> str:
        """Make an identity from the given name.

        If the name is a URL, that can be the identity. Or perhaps it
        needs to start with the default URI prefix (i.e. namespace).

        If the name is not a URL, prefix it with the default URI prefix
        (namespace) and verify that the result is a valid URL.

        We do not support UUIDs, which are legal SBOL identifiers.
        """
        if name is None:
            return None
        name_is_url = SBOLObject._is_url(name)
        if name_is_url:
            return name.strip(posixpath.sep)
        try:
            # If it is a UUID, accept it as the identity
            identity_uuid = uuid.UUID(name)
            return str(identity_uuid)
        except ValueError:
            pass
        # Not a URL or a UUID, so append to the namespace
        base_uri = get_namespace()
        if base_uri.endswith('#'):
            return base_uri + name
        else:
            return posixpath.join(base_uri, name.lstrip(posixpath.sep))

    def validate(self) -> None:
        self._validate_identity()

    @property
    def identity(self) -> str:
        # identity is a read-only property
        return self._identity

    def identity_is_url(self) -> bool:
        return self._is_url(self.identity)

    def find(self, search_string: str) -> Optional['SBOLObject']:
        if self.identity == search_string:
            return self
        if hasattr(self, 'display_id') and self.display_id == search_string:
            return self
        for obj_list in self._owned_objects.values():
            for obj in obj_list:
                result = obj.find(search_string)
                if result is not None:
                    return result
        return None

    def copy(self, target_doc=None, target_namespace=None):

        new_uri = self.identity

        # If caller specified a target_namespace argument, then copy the object into this
        # new namespace.
        if target_namespace:

            # Map the identity of self into the target namespace
            if hasattr(self, 'identity'):
                old_uri = self.identity
                new_uri = replace_namespace(old_uri, target_namespace, self.getTypeURI())

        new_obj = BUILDER_REGISTER[self.type_uri](identity=new_uri,
                                                  type_uri=self.type_uri)

        # Copy properties
        for property_uri, value_store in self._properties.items():
            new_obj._properties[property_uri] = value_store.copy()

            # TODO:
            # Map into new namespace

        # Assign the new object to the target Document
        if target_doc:
            try:
                target_doc.add(new_obj)
            except TypeError:
                pass  # object is not TopLevel

        # When an object is simply being cloned, the value of wasDerivedFrom should be
        # copied exactly as is from self. However, when copy is being used to generate
        # a new entity, the wasDerivedFrom should point back to self.
        if self.identity == new_obj.identity:
            new_obj.derived_from = self.derived_from
        else:
            new_obj.derived_from = self.identity

        # Copy child objects recursively
        for property_uri, object_list in self._owned_objects.items():
            for o in object_list:
                o_copy = o.copy(target_doc, target_namespace)
                new_obj._owned_objects[property_uri].append(o_copy)
                o_copy.parent = self

        return new_obj


def replace_namespace(old_uri, target_namespace, rdf_type):
    """
    Utility function for mapping an SBOL object's identity into a new namespace. The
    rdf_type is used to map to and from sbol-typed namespaces.
    """

    # Work around an issue where the Document itself is being copied and
    # doesn't have its own URI, so old_uri is None. Return empty string
    # because the identity is not allowed to be None.
    if old_uri is None:
        return ''

    # If the value is an SBOL-typed URI, replace both the namespace and class name
    class_name = parseClassName(rdf_type)
    replacement_target = target_namespace + '/' + class_name

    # If not an sbol typed URI, then just replace the namespace
    if replacement_target not in old_uri:
        replacement_target = target_namespace

    if Config.getOption(ConfigOptions.SBOL_TYPED_URIS):
        # Map into a typed namespace
        replacement = getHomespace() + '/' + class_name
    else:
        # Map into a non-typed namespace
        replacement = getHomespace()

    new_uri = old_uri.replace(replacement_target, replacement)
    if type(old_uri) is URIRef:
        return URIRef(new_uri)
    return new_uri


# Global store for builder methods. Custom SBOL classes
# register their builders in this store
BUILDER_REGISTER: Dict[str, Callable[[str, str], SBOLObject]] = {}



from typing import Union, Any, List, Optional

import rdflib

from . import *


class TextPropertyMixin:

    def from_user(self, value: Any) -> Union[None, rdflib.Literal]:
        if value is None:
            return None
        if not isinstance(value, str):
            raise TypeError(f'Expecting string, got {type(value)}')
        return rdflib.Literal(value)

    def to_user(self, value: Any) -> str:
        return str(value)


class TextSingletonProperty(TextPropertyMixin, SingletonProperty):

    def __init__(self, property_owner: Any, property_uri: str,
                 lower_bound: int, upper_bound: int,
                 validation_rules: Optional[List] = None,
                 initial_value: Optional[str] = None):
        super().__init__(property_owner, property_uri,
                         lower_bound, upper_bound, validation_rules)
        if initial_value:
            self.set(initial_value)


class TextListProperty(TextPropertyMixin, ListProperty):

    def __init__(self, property_owner: Any, property_uri: str,
                 lower_bound: int, upper_bound: int,
                 validation_rules: Optional[List] = None,
                 initial_value: Optional[str] = None):
        super().__init__(property_owner, property_uri,
                         lower_bound, upper_bound, validation_rules)
        if initial_value:
            self.set(initial_value)


def TextProperty(property_owner: Any, property_uri: str,
                 lower_bound: int, upper_bound: Union[int, float],
                 validation_rules: Optional[List] = None,
                 initial_value: Optional[Union[str, List[str]]] = None) -> Property:
    if upper_bound == 1:
        return TextSingletonProperty(property_owner, property_uri,
                                     lower_bound, upper_bound,
                                     validation_rules, initial_value)
    return TextListProperty(property_owner, property_uri,
                            lower_bound, upper_bound,
                            validation_rules, initial_value)



import os
import unittest

import sbol3

MODULE_LOCATION = os.path.dirname(os.path.abspath(__file__))
SBOL3_LOCATION = os.path.join(MODULE_LOCATION, 'SBOLTestSuite', 'SBOL3')


class TestImplementation(unittest.TestCase):

    def test_create(self):
        implementation = sbol3.Implementation('impl1')
        self.assertIsNotNone(implementation)
        self.assertIsNone(implementation.built)

    def test_read_from_file(self):
        test_file = os.path.join(SBOL3_LOCATION, 'entity', 'implementation',
                                 'implementation.nt')
        doc = sbol3.Document()
        doc.read(test_file, sbol3.NTRIPLES)
        search_uri = 'https://sbolstandard.org/examples/impl1'
        implementation = doc.find(search_uri)
        self.assertIsNotNone(implementation)
        self.assertIsInstance(implementation, sbol3.Implementation)
        tetr_uri = 'https://sbolstandard.org/examples/TetR_protein'
        built = tetr_uri
        self.assertCountEqual(built, implementation.built)


if __name__ == '__main__':
    unittest.main()



import unittest

import sbol3


class TestLocalSubComponent(unittest.TestCase):

    def test_create(self):
        types = [sbol3.SBO_DNA]
        lsc = sbol3.LocalSubComponent(types)
        self.assertIsNotNone(lsc)
        self.assertEqual(types, lsc.types)
        self.assertEqual([], lsc.locations)

    def test_validation(self):
        types = []
        with self.assertRaises(sbol3.ValidationError):
            sbol3.LocalSubComponent(types)


if __name__ == '__main__':
    unittest.main()
